"""
--- Day 12: Subterranean Sustainability ---
https://adventofcode.com/2018/day/12
"""
from aocd import data


def parsed(data):
    initial, _, *rest = data.splitlines()
    s0 = initial.split()[-1]
    p0 = s0.index("#")
    s0 = s0.strip(".")
    rules = dict(r.split(" => ") for r in rest)
    return s0, p0, rules


def mutate(s, p0, rules):
    s = "...." + s + "...."
    chunks = [s[i : i + 5] for i in range(len(s))]
    full = "".join([rules.get(c, ".") for c in chunks])
    p0 += full.index("#") - 2
    return full.strip("."), p0


def score(s, p0=0):
    return sum(i for i, v in enumerate(s, start=p0) if v == "#")


def part_a(data, n=20):
    s0, p0, rules = parsed(data)
    for i in range(n):
        s1, p1 = mutate(s0, p0, rules)
        if s1 == s0:
            return score(s1, p0=p1 + n - i - 1)
        s0, p0 = s1, p1
    return score(s0, p0)


def part_b(data):
    return part_a(data, n=50000000000)


print("part a:", part_a(data))
print("part b:", part_b(data))



"""
--- Day 19: An Elephant Named Joseph ---
https://adventofcode.com/2016/day/19
"""
from collections import deque

from aocd import data


def part_a(n):
    elves = deque(range(1, int(n) + 1))
    while elves:
        elves.rotate(-1)
        elf = elves.popleft()
    return elf


def part_b(n):
    n = int(n)
    elves = deque(range(1, n + 1))
    elves.rotate((n + 1) // 2)
    while elves:
        elf = elves.popleft()
        elves.rotate(n // 2 + n // 2 - 1)
        n -= 1
    return elf


print("part a:", part_a(data))
print("part b:", part_b(data))



"""
--- Day 8: I Heard You Like Registers ---
https://adventofcode.com/2017/day/8
"""
from aocd import data


class crazydict(dict):
    def __missing__(self, key):
        if key == "wtf":
            return locals
        if key == "max":
            return max
        self[key] = 0
        return 0


def exe(data):
    data += "\n"
    data = data.replace("\n", " else 0\nΣ = max(Σ, *wtf().values())\n")
    data = data.replace("inc", "+=").replace("dec", "-=")
    d = crazydict()
    exec(data, {}, d)
    b = d.pop("Σ")
    return max(d.values()), b


test_data = """\
b inc 5 if a > 1
a inc 1 if b < 5
c dec -10 if a >= 1
c inc -20 if c == 10"""

assert exe(test_data) == (1, 10)
a, b = exe(data)
print("part a:", a)
print("part b:", b)



"""
--- Day 1: The Tyranny of the Rocket Equation ---
https://adventofcode.com/2019/day/1
"""
from aocd import data
import numpy as np


A = np.fromstring(data, dtype=int, sep="\n")
total = 0
while A.any():
    A = (A // 3 - 2).clip(0)
    if not total:
        print("part a:", A.sum())
    total += A.sum()
print("part b:", total)



"""
--- Day 13: Shuttle Search ---
https://adventofcode.com/2020/day/13
"""
import math
from aocd import data

t0, table = data.splitlines()
t0 = int(t0)
table = {-i: int(m) for i, m in enumerate(table.split(",")) if m != "x"}
bs, buses = zip(*table.items())
wait, bus = min([(-t0 % b, b) for b in buses])
print("part a:", wait * bus)

# chinese remainder theorem...
M = math.lcm(*buses)
ns = [M // m for m in buses]
nis = [pow(n, -1, m) for n, m in zip(ns, buses)]
t = sum(b*n*ni for b, n, ni in zip(bs, ns, nis)) % M
print("part b:", t)



"""
--- Day 10: Monitoring Station ---
https://adventofcode.com/2019/day/10
"""
import logging
from aocd import data
from fractions import Fraction
from collections import defaultdict
from operator import itemgetter
from aoc_wim.zgrid import ZGrid


# logging.basicConfig(format="%(message)s", level=logging.INFO)
log = logging.getLogger(__name__)


def parsed(data):
    zs = ZGrid(data).z(val="#", first=False)
    xys = [(int(z.real), int(z.imag)) for z in zs]
    assert len(xys) == data.count("#")
    return xys


def quadrant(a0, a):
    x0, y0 = a0
    x, y = a
    if y < y0 and x >= x0:
        return 1
    if y >= y0 and x > x0:
        return 2
    if y > y0 and x <= x0:
        return 3
    if y <= y0 and x < x0:
        return 4


def grad(a0, a):
    q = quadrant(a0, a)
    x0, y0 = a0
    x, y = a
    try:
        m = abs(Fraction(y - y0, x - x0))
    except ZeroDivisionError:
        m = float("inf")
    if q % 2:
        m *= -1
    return q, m


def best_monitoring_station(data):
    # returns (x, y), n_asteroids_detected
    gradients = defaultdict(set)
    asteroids = parsed(data)
    for i, a0 in enumerate(asteroids):
        for a in asteroids[i + 1 :]:
            q, g = grad(a, a0)
            gradients[a0].add((q, g))
            gradients[a].add(((q + 2) % 4, -g))
    n_collinear = {k: len(v) for k, v in gradients.items()}
    a, n = max(n_collinear.items(), key=itemgetter(1))
    return a, n


def norm2(a1, a2):
    dx = a2[0] - a1[0]
    dy = a2[1] - a1[1]
    d2 = dx * dx + dy * dy
    return d2


def part_ab(data, target=200):
    n_max = data.count("#") - 1
    target = min(target, n_max)
    asteroids = parsed(data)
    a0, part_a_answer = best_monitoring_station(data)
    d = defaultdict(list)
    for a in asteroids:
        if a == a0:
            continue
        g = grad(a0, a)
        d[g].append(a)
    ds = {}
    for g in sorted(d):
        ds[g] = sorted(d[g], key=lambda a: norm2(a0, a), reverse=True)
    i = 0
    while True:
        for g in ds:
            if ds[g]:
                a = ds[g].pop()
                i += 1
                s = {1: "st", 2: "nd", 3: "rd"}.get(i % 10, "th")
                f = "and final " if i == n_max else ""
                log.info("The %d%s %sasteroid to be vaporized is at %d,%d.", i, s, f, *a)
                if i == target:
                    part_b_answer = 100 * a[0] + a[1]
                    return part_a_answer, part_b_answer


if __name__ == "__main__":
    a, b = part_ab(data)
    print("part a:", a)
    print("part b:", b)



"""
--- Day 11: Seating System ---
https://adventofcode.com/2020/day/11
"""
from aocd import data
from aoc_wim.zgrid import ZGrid


def evolve(grid, part="a"):
    result = ZGrid(grid.d.copy())
    max_occ = 4 if part == "a" else 5
    for z0, c0 in grid.items():
        if c0 == ".":
            continue  # floor
        if part == "a":
            n_occ = grid.count_near(z0, "#", n=8)
        else:
            n_occ = 0
            for dz in grid.near(0, n=8):
                z = z0
                val = "."
                while val == ".":
                    z += dz
                    val = grid.get(z)
                    n_occ += val == "#"
        if c0 == "L" and n_occ == 0:
            result[z0] = "#"
        if c0 == "#" and n_occ >= max_occ:
            result[z0] = "L"
    return result


for p in "ab":
    grid0 = ZGrid(data)
    while True:
        grid1 = evolve(grid0, part=p)
        grid1.draw(clear=True, pretty=True, transform={".": "  ", "L": "💺", "#": "🧘🏽"})
        if grid1.d == grid0.d:
            break
        grid0 = grid1
    print("part", p, grid0.count("#"))



"""
--- Day 12: Rain Risk ---
https://adventofcode.com/2020/day/12
"""
from aocd import data
from aoc_wim.zgrid import ZGrid
from aoc_wim.zgrid import manhattan_distance

za = zb = 0                          # ferry
dza = ZGrid.east                     # direction (part a)
dzb = 10 * ZGrid.east + ZGrid.north  # velocity (part b)

for line in data.splitlines():
    i = line[0]
    n = int(line[1:])
    if i in "NSEW":
        za += n * getattr(ZGrid, i)
        dzb += n * getattr(ZGrid, i)
    elif i in "LR":
        n //= 90
        turn = getattr(ZGrid, f"turn{i}")
        dza *= turn ** n
        dzb *= turn ** n
    elif i == "F":
        za += n * dza
        zb += n * dzb

print("part a:", manhattan_distance(za))
print("part b:", manhattan_distance(zb))



"""
--- Day 12: Passage Pathing ---
https://adventofcode.com/2021/day/12
"""
from aocd import data
from collections import defaultdict
from collections import deque


caves = defaultdict(set)
for line in data.splitlines():
    x, y = line.split("-")
    caves[x].add(y)
    caves[y].add(x)


def get_neighbors_a(here, visited):
    return [c for c in caves[here] if c.isupper() or c not in visited]


def get_neighbors_b(here, visited):
    if any(v > 1 for k, v in visited.items() if not k.isupper()):
        # we've already visited a small cave twice
        return get_neighbors_a(here, visited)
    return [n for n in caves[here] if n != "start"]


def n_paths(part):
    get_neighbors = get_neighbors_a if part == "a" else get_neighbors_b
    result = 0
    q = deque([(["start"], {"start": 1})])
    while q:
        path, visited = q.pop()
        here = path[-1]
        if here == "end":
            result += 1
            continue
        neighbors = get_neighbors(here, visited)
        q.extend([(path + [n], {**visited, n: visited.get(n, 0) + 1}) for n in neighbors])
    return result


print("part a:", n_paths("a"))
print("part b:", n_paths("b"))



"""
--- Day 22: Slam Shuffle ---
https://adventofcode.com/2019/day/22
"""
from math import prod
from aocd import data


def parsed(data):
    # decompose "flip" operations into cut + stride
    data = data.replace("deal into new stack", "cut -1\ndeal with increment -1")
    ops = []
    for line in data.splitlines():
        op, *crap, n = line.split()
        n = int(n)
        assert op in {"cut", "deal"}
        ops.append((op, n))
    return ops


def simplify(ops, N, repeats=1):
    done = False
    while not done:
        done = True
        for i in range(1, len(ops)):
            op_prev, n_prev = ops[i - 1]
            op, n = ops[i]
            if op == "cut" and op_prev == "deal":
                # bubble all the cuts up to the top
                n_cut = (n * pow(n_prev, N - 2, N)) % N
                ops[i] = op_prev, n_prev
                ops[i - 1] = op, n_cut
                done = False
    # merge all the rotations
    n_cut = sum(n for op, n in ops if op == "cut") % N
    # merge all the strides
    n_deal = prod(n for op, n in ops if op == "deal") % N
    ops[:] = [("cut", n_cut), ("deal", n_deal)]
    if repeats > 1:
        compress(ops, N, repeats)


def compress(ops, N, repeats):
    assert len(ops) == 2
    result = []
    while repeats > 1:
        if repeats % 2:
            result = ops + result
            repeats -= 1
        repeats //= 2
        ops *= 2
        simplify(ops, N)
    result = ops + result
    simplify(result, N)
    ops[:] = result


def part_a(data, N=10007, pos=2019):
    ops = parsed(data)
    simplify(ops, N)
    [(cut, c0), (deal, f)] = ops
    result = f * (pos - c0) % N
    return result


def part_b(data, N=119315717514047, n_shuffles=101741582076661, card=2020):
    ops = parsed(data)
    simplify(ops, N, n_shuffles)
    [(cut, c0), (deal, f)] = ops
    # fermat's little theorem - modular multiplicative inverse
    # trick works because N is prime (!)
    result = (c0 + card * pow(f, N - 2, N)) % N
    return result


print("part a", part_a(data))
print("part b", part_b(data))



"""
--- Day 9: Smoke Basin ---
https://adventofcode.com/2021/day/9
"""
from aocd import data
import networkx as nx
from aoc_wim.zgrid import ZGrid

g = ZGrid(data)
bs = sorted(nx.connected_components(g.graph(extra="012345678")), key=len)
print("part a:", sum(min([1 + int(g[z]) for z in b]) for b in bs))
print("part b:", len(bs[-1]) * len(bs[-2]) * len(bs[-3]))



"""
--- Day 13: Shuttle Search ---
https://adventofcode.com/2020/day/13
"""
from aocd import data

t0, table = data.splitlines()
t0 = int(t0)
buses = [(i, int(b)) for i, b in enumerate(table.split(",")) if b != "x"]
wait, bus = min([(-t0 % b, b) for _, b in buses])
print("part a:", wait * bus)

timestamp = 0
increment = 1
synced = {}
while len(synced) < len(buses):
    timestamp += increment
    for phase, bus in buses:
        if bus not in synced:
            if (timestamp + phase) % bus == 0:
                increment *= bus
                synced[bus] = timestamp
print("part b:", timestamp)



"""
--- Day 20: Firewall Rules ---
https://adventofcode.com/2016/day/20
"""
from aocd import data


def cleanup_data(data):
    intervals = []
    for line in data.splitlines():
        left, right = line.split("-")
        interval = int(left), int(right)
        intervals.append(interval)
    intervals.sort()

    while True:
        n = len(intervals)
        for i in range(n - 1):
            lo_p, hi_p = intervals[i]
            lo, hi = intervals[i + 1]
            if lo <= hi_p + 1:
                del intervals[i + 1]
                intervals[i] = lo_p, max(hi_p, hi)
                break
        else:
            break

    return intervals


def part_a(clean_data):
    return clean_data[0][1] + 1


def part_b(clean_data, n_min=0, n_max=4294967295):
    n = n_max - n_min + 1
    for lo, high in clean_data:
        n -= high - lo + 1
    return n


if __name__ == "__main__":
    clean_data = cleanup_data(data)
    print(part_a(clean_data))
    print(part_b(clean_data))



"""
--- Day 5: Alchemical Reduction ---
https://adventofcode.com/2018/day/5
"""
from aocd import data


def react(data, remove=None):
    if remove is not None:
        data = data.replace(remove, "").replace(remove.swapcase(), "")
    result = []
    for char in data:
        if result and char == result[-1].swapcase():
            result.pop()
        else:
            result.append(char)
    return "".join(result)


print("part a:", len(react(data)))


def choices(data):
    units = set(data.lower())
    results = {u: react(data, remove=u) for u in units}
    return results


print("part b:", min([len(v) for v in choices(data).values()], default=0))



"""
--- Day 7: No Space Left On Device ---
https://adventofcode.com/2022/day/7
"""
from aocd import data

lines = iter(data.splitlines())
assert next(lines) == "$ cd /"
cwd = "/"
dirs = {"/": 0}  # map of dirs to their cumulative size
parents = []  # cwd stack
for line in lines:
    match line.split():
        case ["$", "cd", ".."]:
            cwd = parents.pop()
        case ["$", "cd", name]:
            parents.append(cwd)
            cwd += name + "/"
        case ["$", "ls"]:
            pass
        case ["dir", name]:
            dirs[cwd + name + "/"] = 0
        case [s, name]:
            size = int(s)
            for d in parents + [cwd]:
                dirs[d] += size

total_space = 70_000_000
df_target = 30_000_000
used_space = dirs["/"]
df_actual = total_space - used_space
del_needed = df_target - df_actual

a = sum(v for v in dirs.values() if v <= 100_000)
b = min(v for v in dirs.values() if v >= del_needed)

print("part a:", a)
print("part b:", b)



"""
--- Day 3: Spiral Memory ---
https://adventofcode.com/2017/day/3
"""
from aocd import data
from aoc_wim.zgrid import manhattan_distance


def gen():
    # yields pairs of (manhattan distance from origin, sum of neighbours)
    pos, delta = 0, 1
    sums = {pos: 1}
    yield pos, sums[pos]
    while True:
        pos += 1 - 1j
        for _ in range(4):
            delta *= 1j
            for _ in range(int(2 * abs(pos.real))):
                pos += delta
                neighbours = (pos + dr + dj for dr in (-1, 0, 1) for dj in (-1j, 0, 1j))
                sums[pos] = sum(sums.get(p, 0) for p in neighbours)
                yield manhattan_distance(pos), sums[pos]


def part_ab(data):
    n = int(data)
    a = b = None
    for i, (distance, sum_) in enumerate(gen(), 1):
        if i == n:
            a = a or distance
        if sum_ > n:
            b = b or sum_
        if a is not None and b is not None:
            return a, b


if __name__ == "__main__":
    a, b = part_ab(data)
    print("part a:", a)
    print("part b:", b)



"""
--- Day 12: JSAbacusFramework.io ---
https://adventofcode.com/2015/day/12
"""
import json
import re

from aocd import data


def sum_of_numbers_in_text(s):
    return sum(int(n) for n in re.findall(r"-?\d+", s))


def rsum(data):
    if isinstance(data, int):
        return data
    elif isinstance(data, dict):
        if "red" in data.values():
            return 0
        return rsum(list(data.values()))
    elif isinstance(data, list):
        return sum(rsum(n) for n in data)
    else:
        return 0


print(sum_of_numbers_in_text(data))
print(rsum(json.loads(data)))



"""
--- Day 19: Monster Messages ---
https://adventofcode.com/2020/day/19
"""
from aocd import data
import lark


def solve(rules, messages):
    rules = rules.translate(str.maketrans("0123456789", "abcdefghij"))
    parser = lark.Lark(rules, start="a")
    result = 0
    for message in messages.splitlines():
        try:
            parser.parse(message)
        except lark.LarkError:
            pass
        else:
            result += 1
    return result


rules, messages = data.split("\n\n")
print("part a:", solve(rules, messages))

rules = rules.replace("8: 42", "8: 42 | 42 8").replace("11: 42 31", "11: 42 31 | 42 11 31")
print("part b:", solve(rules, messages))



"""
--- Day 3: Binary Diagnostic ---
https://adventofcode.com/2021/day/3
"""
from aocd import data


rows = data.splitlines()
cols = ["".join(c) for c in zip(*rows)]
h2 = len(rows) // 2
n = "".join("10"[c.count("1") >= h2] for c in cols)
𝛾 = int(n, 2)
ε = 𝛾 ^ 2 ** len(n) - 1
print("part a:", 𝛾 * ε)

life_support = 1
s = "01"
for _ in s:
    i = 0
    ns = rows.copy()
    while len(ns) > 1:
        col = [n[i] for n in ns]
        b = col.count("1") >= col.count("0")
        ns = [n for n in ns if n[i] == s[b]]
        i += 1
    [n] = ns
    life_support *= int(n, 2)
    s = s[::-1]
print("part b:", life_support)



"""
--- Day 3: Crossed Wires ---
https://adventofcode.com/2019/day/3
"""
from aocd import data
from aoc_wim.zgrid import manhattan_distance
from aoc_wim.zgrid import ZGrid


wires = data.splitlines()
seen = []
for i, wire in enumerate(wires):
    seen.append({})
    z = d = 0
    for step in wire.split(","):
        dz = getattr(ZGrid, step[0])
        for _ in range(int(step[1:])):
            z += dz
            d += 1
            if z not in seen[-1]:
                seen[-1][z] = d

crossings = seen[0].keys() & seen[1].keys()
print("part a:", min([manhattan_distance(z) for z in crossings]))
print("part b:", min([seen[0][z] + seen[1][z] for z in crossings]))



"""
--- Day 7: Internet Protocol Version 7 ---
https://adventofcode.com/2016/day/7
"""
import re

from aocd import data


def splitter(s):
    pattern = r"\[([a-z]*)\]"
    ins = re.findall(pattern, s)
    outs = re.sub(pattern, " ", s).split()
    return ins, outs


def chunker(s, size):
    i = 0
    while True:
        chunk = s[i : i + size]
        if len(chunk) != size:
            break
        yield chunk
        i += 1


def has_abba(s):
    for a, b, b_, a_ in chunker(s, 4):
        if a == a_ and b == b_ and a != b:
            return True
    return False


def gen_aba(s):
    for a, b, a_ in chunker(s, 3):
        if a == a_ and a != b:
            yield a + b + a


def has_aba(s):
    return bool(next(gen_aba(s), False))


def support_tls(s):
    ins, outs = splitter(s)
    return has_abba(".".join(outs)) and not has_abba(".".join(ins))


def support_ssl(s):
    ins, outs = splitter(s)
    for a, b, a in gen_aba("..".join(ins)):
        bab = b + a + b
        if bab in "..".join(outs):
            return True
    return False


tls = ssl = 0
for line in data.splitlines():
    tls += support_tls(line)
    ssl += support_ssl(line)

print(tls)
print(ssl)



"""
--- Day 9: Encoding Error ---
https://adventofcode.com/2020/day/9
"""
from aocd import data
from wimpy import chunks
from aoc_wim.aoc2020 import find_pair

numbers = [int(x) for x in data.splitlines()]
n = 25
if len(numbers) < n:
    n = 5  # hackish - lower "preamble" for the example test to work

for *pre, target in chunks(numbers, chunk_size=n + 1, overlap=n):
    if find_pair(set(pre), target) is None:
        print("part a:", target)
        break

start = 0
stop = 2
total = sum(numbers[start:stop])
while total != target or start == stop - 1:
    if total < target:
        total += numbers[stop]
        stop += 1
    else:
        total -= numbers[start]
        start += 1
chunk = numbers[start:stop]
assert sum(chunk) == target, "the chunk must sum to the target value"
assert len(chunk) >= 2, "the chunk must be a contiguous set of at least two numbers"
print("part b:", min(chunk) + max(chunk))
